import rich_click as click
from .build import build
from .publish import publish
from .stage import stage
from .init import init
from .wizard import wizard
from .validate import validate
from click import Context


@click.group()
@click.pass_context
def kabinet(ctx: Context) -> None:
    """Deploy the arkitekt_next app with Port

    The port deployer is an arkitekt_next plugin service, which allows you to deploy your arkitekt_next app to
    any arkitekt_next instance and make it instantly available to the world. Port uses docker to containerize
    your application and will publish it locally to your dockerhub account, and mark it locally as
    deployed. People can then use your github repository to deploy your app to their arkitekt_next instance.

    """

    pass


kabinet.add_command(build, "build")
kabinet.add_command(init, "init")
kabinet.add_command(validate, "validate")
kabinet.add_command(publish, "publish")
kabinet.add_command(stage, "stage")
kabinet.add_command(wizard, "wizard")



import os
import rich_click as click
from arkitekt_next.cli.types import Requirement, Framework
from rich.table import Table
from rich.panel import Panel
from rich.console import Group
from arkitekt_next.cli.vars import get_console


def check_dl_frameworks():
    import site
    import sys

    site_packages = site.getsitepackages()
    site_packages.append(sys.prefix)

    frameworks = {
        "tensorflow": Framework.TENSORFLOW,
        "torch": Framework.PYTORCH,
        "torchvision": Framework.PYTORCH,
        "tensorflow-gpu": Framework.TENSORFLOW,
        "torch-gpu": Framework.PYTORCH,
        "torchvision-gpu": Framework.PYTORCH,
    }

    included_frameworks = set()

    for key, value in frameworks.items():
        for site_package in site_packages:
            if os.path.exists(os.path.join(site_package, key)):
                included_frameworks.add(value)

    return included_frameworks


def inspect_requirements(automatic=False):
    """Inspect the requirements and


    return a list of requirements and reasons.

    Parameters
    ----------
    automatic : bool, optional
        Should we automically add all of the requirements, by default False

    Returns
    -------
    _type_
        _description_
    """
    requirements = []
    reasons = []
    frameworks = check_dl_frameworks()
    if frameworks:
        requirements.append(Requirement.GPU)
        reasons.append(
            "Deep learning framework detected ('{}')".format(", ".join(frameworks))
        )

    return requirements, reasons


@click.command()
@click.pass_context
def wizard(ctx) -> None:
    """Inspect the current project for requirements

    Inspects the current project and provides a list of [i]suggest[/]ed requirements
    that should be added to the manifest. You should run this command in the same
    virtual environment as you are running your app, as it will detect the installed
    packages and provide a list of requirements that should be added to the manifest.

    """

    console = get_console(ctx)
    requirements, reasons = inspect_requirements()

    table = Table.grid()
    table.padding = (0, 4, 0, 0)
    table.add_column("Requirement")
    table.add_column("Reason")

    for item in zip(requirements, reasons):
        table.add_row(str(item[0].value), item[1])

    panel = Panel(
        Group("[bold green]Manifest[/]", table),
        title_align="center",
        border_style="green",
        style="white",
    )

    console.print(panel)



from pydantic import BaseModel, Field, validator
import datetime
from typing import List, Optional, Union, Literal, Dict
from enum import Enum
import semver
import uuid
from arkitekt_next.model import Requirement
from rekuest.api.schema import DefinitionInput
from string import Formatter
import os

from rekuest_next.api.schema import TemplateInput

ALLOWED_BUILDER_KEYS = [
    "tag",
    "dockerfile",
]


class Framework(str, Enum):
    """Do we support other frameworks?"""

    VANILLA = "vanilla"
    TENSORFLOW = "tensorflow"
    PYTORCH = "pytorch"


class Packager(str, Enum):
    CONDA = "conda"
    POETRY = "poetry"
    PIP = "pip"


class Manifest(BaseModel):
    identifier: str
    version: str
    author: str
    logo: Optional[str]
    entrypoint: str
    scopes: List[str]
    created_at: datetime.datetime = Field(default_factory=datetime.datetime.now)

    @validator("version", pre=True)
    def version_must_be_semver(cls, v) -> str:
        """Checks that the version is a valid semver version"""
        if isinstance(v, str):
            try:
                semver.VersionInfo.parse(v)
            except ValueError:
                raise ValueError("Version must be a valid semver version")
        return str(v)

    def to_console_string(self):
        return f"📦 {self.identifier} ({self.version}) by {self.author}"

    def to_builder_dict(self):
        return {
            "identifier": self.identifier,
            "version": self.version,
            "logo": self.logo,
            "scopes": self.scopes,
        }

    class Config:
        validate_assignment = True


class SelectorType(str, Enum):
    RAM = "ram"
    CPU = "cpu"
    GPU = "gpu"
    LABEL = "label"


class BaseSelector(BaseModel):
    """A selector is a way to describe a flavours preference for a
    compute node. It contains the node_id, the selector and the flavour_id.
    """

    required: bool = True

    class Config:
        extra = "forbid"

    def build_docker_params(self) -> List[str]:
        """Builds the docker params for this selector

        Should return a list of strings that can be used as docker params
        If the selector is not required, it should return an empty list

        Returns
        -------
        List[str]
            The docker params for this selector
        """
        return []

    def build_arkitekt_next_params(self) -> List[str]:
        """Builds the arkitekt_next params for this selector

        Returns
        -------
        List[str]
            The docker params for this selector
        """
        return []


class RAMSelector(BaseSelector):
    """A selector is a way to describe a flavours preference for a
    compute node. It contains the node_id, the selector and the flavour_id.
    """

    type: Literal["ram"]
    min: int


class CPUSelector(BaseSelector):
    """A selector is a way to describe a flavours preference for a
    compute node. It contains the node_id, the selector and the flavour_id.
    """

    type: Literal["cpu"]
    min: int
    frequency: Optional[int] = None


class CudaSelector(BaseSelector):
    """A selector is a way to describe a flavours preference for a
    compute node. It contains the node_id, the selector and the flavour_id.
    """

    type: Literal["cuda"]
    frequency: Optional[int] = Field(default=None, description="The frequency in MHz")
    memory: Optional[int] = Field(default=None, description="The memory in MB")
    architecture: Optional[str] = Field(
        default=None, description="The architecture of the GPU"
    )
    compute_capability: str = Field(
        default="3.5", description="The minimum compute capability"
    )
    cuda_cores: Optional[int] = None
    cuda_version: str = Field(default="10.2", description="The minimum cuda version")

    def build_docker_params(self) -> List[str]:
        """Builds the docker params for this selector

        Should return a list of strings that can be used as docker params
        If the selector is not required, it should return an empty list

        Returns
        -------
        List[str]
            The docker params for this selector
        """
        return [
            "--gpus",
            "all",
        ]


class RocmSelector(BaseSelector):
    """A selector is a way to describe a flavours preference for a
    compute node. It contains the node_id, the selector and the flavour_id.
    """

    type: Literal["rocm"]
    min: int
    frequency: Optional[int] = None
    memory: Optional[int] = None
    architecture: Optional[str] = None
    compute_capability: Optional[str] = None
    cuda_cores: Optional[int] = None
    cuda_version: Optional[str] = None

    def build_docker_params(self) -> List[str]:
        """Builds the docker params for this selector"""

        return [
            "--device=/dev/kfd",
            "--device=/dev/dri",
            "--group-add",
            "video",
            "--cap-add=SYS_PTRACE",
            "--security-opt",
            "seccomp=unconfined",
        ]


class LabelSelector(BaseSelector):
    """A selector is a way to describe a flavours preference for a
    compute node. It contains the node_id, the selector and the flavour_id.
    """

    type: Literal["label"]
    key: str
    value: str


class ServiceSelector(BaseSelector):
    """A service selector is a way to describe a flavours preference for a
    service. It contains the service_id,
    """

    type: Literal["service"]


Selector = Union[
    RAMSelector, CPUSelector, CudaSelector, RocmSelector, LabelSelector, ServiceSelector
]


class Inspection(BaseModel):
    templates: List[TemplateInput]
    requirements: Dict[str, Requirement]
    size: int


class Flavour(BaseModel):
    selectors: List[Selector]
    description: str = Field(default="")
    dockerfile: str = Field(default="Dockerfile")
    build_command: List[str] = Field(
        default_factory=lambda: [
            "docker",
            "build",
            "-t",
            "{tag}",
            "-f",
            "{dockerfile}",
            ".",
        ]
    )

    @validator("build_command", each_item=True, always=True)
    def check_valid_template_name(cls, v):
        """Checks that the template name is valid"""
        for literal_text, field_name, format_spec, conversion in Formatter().parse(v):
            if field_name is not None:
                assert (
                    field_name in ALLOWED_BUILDER_KEYS
                ), f"Invalid template key {field_name}. Allowed keys are {ALLOWED_BUILDER_KEYS}"

        return v

    def generate_build_command(self, tag: str, relative_dir: str):
        """Generates the build command for this flavour"""

        dockerfile = os.path.join(relative_dir, self.dockerfile)

        return [v.format(tag=tag, dockerfile=dockerfile) for v in self.build_command]

    def check_relative_paths(self, flavour_folder: str):
        """Checks that the paths are relative to the flavour folder"""

        dockerfile_path = os.path.join(flavour_folder, self.dockerfile)

        if not os.path.exists(dockerfile_path):
            raise Exception(
                f"Could not find Dockerfile {self.dockerfile} in flavour {flavour_folder}"
            )


class Deployment(BaseModel):
    """A deployment is a Release of a Build.
    It contains the build_id, the manifest, the builder, the definitions, the image and the deployed_at timestamp.



    """

    deployment_run: str = Field(
        default_factory=lambda: str(uuid.uuid4()),
        description="The unique identifier of the deployment run",
    )

    deployment_id: str = Field(
        default_factory=lambda: str(uuid.uuid4()),
        description="The unique identifier of the deployment",
    )
    manifest: Manifest = Field(description="The manifest of the app that was deployed")
    selectors: List[Selector] = Field(
        description="The selectors are used to place this image on the nodes",
        default_factory=list,
    )
    flavour: str = Field(
        description="The flavour that was used to build this deployment",
        default="vanilla",
    )
    build_id: str = Field(
        description="The build_id of the build that was deployed. Is referenced in the build.yaml file."
    )
    inspection: Optional[Inspection] = Field(
        description="The inspection of the build that was deployed"
    )
    image: str = Field(
        description="The docker image that was built for this deployment"
    )
    deployed_at: datetime.datetime = Field(
        default_factory=datetime.datetime.now,
        description="The timestamp of the deployment",
    )


class DeploymentsConfigFile(BaseModel):
    """The ConfigFile is a pydantic model that represents the deployments.yaml file


    Parameters
    ----------
    BaseModel : _type_
        _description_
    """

    deployments: List[Deployment] = []
    latest_deployment_run: Optional[str] = None


class Build(BaseModel):
    build_run: str
    build_id: str
    inspection: Optional[Inspection] = None
    description: str = Field(default="")
    selectors: List[Selector] = Field(default_factory=list)
    flavour: str = Field(default="vanilla")
    manifest: Manifest
    build_at: datetime.datetime = Field(default_factory=datetime.datetime.now)
    base_docker_command: List[str] = Field(
        default_factory=lambda: ["docker", "run", "-it", "--net", "host"]
    )
    base_arkitekt_next_command: List[str] = Field(
        default_factory=lambda: ["arkitekt-next", "run", "prod", "--headless"]
    )

    def build_docker_command(self) -> List[str]:
        """Builds the docker command for this build"""

        base_command = self.base_docker_command

        for selector in self.selectors:
            base_command = base_command + selector.build_docker_params()

        base_command = base_command + [self.build_id]

        return base_command

    def build_arkitekt_next_command(self, fakts_url: str):
        """Builds the arkitekt_next command for this build"""

        base_command = self.base_arkitekt_next_command

        for selector in self.selectors:
            base_command = base_command + selector.build_arkitekt_next_params()

        base_command = base_command + ["--url", fakts_url]

        return base_command


class BuildsConfigFile(BaseModel):
    builds: List[Build] = Field(default_factory=list)
    latest_build_run: Optional[str] = None



import click

from arkitekt_next.bloks.services.config import ConfigService
from arkitekt_next.bloks.services.gateway import GatewayService
from arkitekt_next.bloks.services.name import NameService
from arkitekt_next.bloks.services.secret import SecretService
from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import YamlFile, Repo
from pydantic import BaseModel
from typing import Dict, Optional
import secrets
from blok import blok, InitContext
from arkitekt_next.bloks.services.s3 import S3Credentials, S3Service


class BucketMapParamType(click.ParamType):
    name = "redeem_token"

    def convert(self, value, param, ctx):
        if isinstance(value, dict):
            return value
        try:
            user, token = value.split(":")
            return {"user": user, "token": token}
        except ValueError:
            self.fail(
                f"RedeemToken '{value}' is not in the correct format. It should be 'username:token'.",
                param,
                ctx,
            )


TOKEN = BucketMapParamType()


@blok(S3Service)
class MinioBlok:
    db_name: str

    def __init__(self) -> None:
        self.users = []
        self.username = secrets.token_hex(16)
        self.password = secrets.token_hex(16)
        self.protocol = "http"
        self.host = "minio"
        self.port = 9000
        self.skip = False
        self.scopes = {}
        self.init_image = "jhnnsrs/init:paper"
        self.minio_image = "minio/minio:RELEASE.2023-02-10T18-48-39Z"
        self.buckets = []
        self.registered_clients = []
        self.preformed_bucket_names = [secrets.token_hex(16) for i in range(100)]
        self.preformed_access_keys = [secrets.token_hex(16) for i in range(100)]
        self.preformed_secret_keys = [secrets.token_hex(16) for i in range(100)]

    def create_buckets(self, buckets: list[str]) -> S3Credentials:
        new_access_key = self.secret_service.retrieve_secret()
        new_secret_key = self.secret_service.retrieve_secret()

        bucket_map = {}

        for bucket in buckets:
            bucket_name = self.name_service.retrieve_name()
            bucket_map[bucket] = bucket_name
            self.gateway_service.expose(bucket_name, 9000, self.host)

        self.buckets.extend(bucket_map.values())

        creds = S3Credentials(
            name=self.name_service.retrieve_name(),
            access_key=new_access_key,
            buckets=bucket_map,
            host=self.host,
            port=self.port,
            secret_key=new_secret_key,
            protocol=self.protocol,
            dependency=self.host if not self.skip else None,
        )

        self.registered_clients.append(creds)

        return creds

    def preflight(
        self,
        init: InitContext,
        secret: SecretService,
        name: NameService,
        gateway: GatewayService,
        config: ConfigService,
    ):
        for key, value in init.kwargs.items():
            setattr(self, key, value)

        self.secret_service = secret
        self.name_service = name
        self.gateway_service = gateway

        self.config_path = config.get_path(self.host + ".yaml")

    def build(self, context: ExecutionContext):
        minio_service_init = {
            "depends_on": {
                "minio": {
                    "condition": "service_started",
                },
            },
            "environment": {
                "MINIO_ROOT_PASSWORD": self.password,
                "MINIO_ROOT_USER": self.username,
                "MINIO_HOST": f"http://{self.host}:9000",
            },
            "image": self.init_image,
            "volumes": [f"{self.config_path}:/workspace/config.yaml"],
        }

        minio_service = {
            "command": ["server", "/data"],
            "environment": {
                "MINIO_ROOT_PASSWORD": self.password,
                "MINIO_ROOT_USER": self.username,
            },
            "image": self.minio_image,
            "volumes": ["./data:/data"],
            "labels": ["fakts.service=live.arkitekt.s3", "fakts.builder=arkitekt.s3"],
        }

        context.file_tree.set_nested("data", {})

        context.docker_compose.set_nested("services", self.host, minio_service)
        context.docker_compose.set_nested(
            "services", f"{self.host}_init", minio_service_init
        )

        context.file_tree.set_nested(
            *self.config_path.split("/"),
            YamlFile(
                buckets=[{"name": i} for i in self.buckets],
                users=[
                    {
                        "access_key": i.access_key,
                        "secret_key": i.secret_key,
                        "name": i.name,
                        "policies": ["readwrite"],
                    }
                    for i in self.registered_clients
                ],
            ),
        )

    def get_options(self):
        with_host = Option(
            subcommand="host",
            help="The fakts url for connection",
            default=self.host,
        )
        with_username = Option(
            subcommand="username",
            help="The fakts url for connection",
            default=self.username,
        )
        with_password = Option(
            subcommand="password",
            help="The fakts url for connection",
            default=self.password,
        )

        return [
            with_host,
            with_password,
            with_username,
        ]



from typing import Dict, Any, Protocol
from blok import blok, InitContext, Option, ExecutionContext, service
from blok.tree import YamlFile




@service("live.arkitekt.config")
class ConfigService(Protocol):

    def register_config(self, name: str, file: YamlFile) -> str:
        ...

    def get_path(self, name: str) -> str:
        ...
        return f"./{self.config_path}/" + name

   



import rich_click as click
from semver import (
    parse as parse_semver,
    bump_build,
    bump_major,
    bump_minor,
    bump_patch,
    bump_prerelease,
)
from arkitekt_next.cli.vars import get_console, get_manifest
from arkitekt_next.cli.io import write_manifest


@click.group()
@click.pass_context
def version(ctx):
    """Updates the version of the arkitekt_next app

    ArkitektNext manifests versioning follow [link=https://semver.org]semver[/link] and are used to version the app.
    This provides an orthogonal way to version the app, beyond node versioning. The version is used to
    track changes and to provide a way to update the app in the platform. For more information, please visit
    [link=https://arkitekt_next.live]https://arkitekt_next.live[/link]

    """


@version.command("set")
@click.argument("VERSION", type=str, required=False)
@click.pass_context
def set_version(ctx, version):
    """Sets the version of the arkitekt_next app

    When setting the version, you can either provide a version, or you can let the cli
    prompt you for a version. If you provide a version, it will be parsed and validated
    against semver. If you let the cli prompt you, it will try to parse the current version
    and suggest a new version based on that. If the current version is not a valid semver
    version, it will prompt you for a new version without a suggestion.
    """

    manifest = get_manifest(ctx)
    get_console(ctx)
    old_version = manifest.version

    if not version:
        try:
            potential_new_version = bump_patch(old_version)
        except Exception:
            potential_new_version = None

        new_version = click.prompt(
            "Please provide a new version", default=potential_new_version, type=str
        )
        version = parse_semver(new_version)
        version = new_version

    manifest.version = version
    write_manifest(manifest)
    get_console().print(f"Version Updated from {old_version} to {version}")


@version.command()
@click.pass_context
def patch(ctx):
    """ "Patches the version of the arkitekt_next app


    Patches the version of the arkitekt_next app, by bumping the patch number.
    E.g. from 1.0.1 to 1.0.2. This should be used for bugfixes and small changes.
    """
    manifest = get_manifest(ctx)
    console = get_console(ctx)
    old_version = manifest.version
    manifest.version = bump_patch(old_version)
    write_manifest(manifest)
    console.print(f"Version Updated from {old_version} to {manifest.version}")


@version.command()
@click.pass_context
def minor(ctx):
    """Bumps the minor version number of the arkitekt_next app

    Patches the version of the arkitekt_next app, by bumping the minor number.
    E.g. from 1.0.1 to 1.1.1. This should be used for new features, that
    are backwards compatible.

    """
    manifest = get_manifest(ctx)
    console = get_console(ctx)
    old_version = manifest.version
    manifest.version = bump_minor(old_version)
    write_manifest(manifest)
    console.print(f"Version Updated from {old_version} to {manifest.version}")


@version.command()
@click.pass_context
def major(ctx):
    """Increase the major version of the arkitekt_next app"

    Patches the version of the arkitekt_next app, by bumping the major number.
    E.g. from 1.0.1 to 2.0.1, This should be used for breaking changes,
    that are not backwards compatible (e.g. deleting a node).

    """
    manifest = get_manifest(ctx)
    console = get_console(ctx)
    old_version = manifest.version
    manifest.version = bump_major(old_version)
    write_manifest(manifest)
    console.print(f"Version Updated from {old_version} to {manifest.version}")


@version.command()
@click.pass_context
def prerelease(ctx):
    """Patches the prerelease of the arkitekt_next app"


    Patches the version of the arkitekt_next app, by bumping the prerelease number.
    E.g. from 1.0.1 to 1.0.1-alpha.1
    """
    manifest = get_manifest(ctx)
    console = get_console(ctx)
    old_version = manifest.version
    manifest.version = bump_prerelease(old_version)
    write_manifest(manifest)
    console.print(f"Version Updated from {old_version} to {manifest.version}")


@version.command("build")
@click.pass_context
def bump_build(ctx):
    """Patches the build of the arkitekt_next app

    Patches the version of the arkitekt_next app, by bumping the build number.
    E.g. from 1.0.1 to 1.0.1+1, This should be used for changes that are not
    reflected in the version number, but are still important to track (e.g. a
    hotfix).

    """
    manifest = get_manifest(ctx)
    console = get_console(ctx)
    old_version = manifest.version
    manifest.version = bump_build(old_version)
    write_manifest(manifest)
    console.print(f"Version Updated from {old_version} to {manifest.version}")



import click
from pydantic import BaseModel
from typing import Dict, Any
import yaml
import secrets
from blok import blok, InitContext

from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import YamlFile, Repo
from arkitekt_next.bloks.funcs import (
    create_default_service_dependencies,
    create_default_service_yaml,
    DefaultService,
)


class AccessCredentials(BaseModel):
    password: str
    username: str
    host: str
    port: str
    db_name: str


@blok("live.arkitekt.mikro")
class MikroBlok:
    def __init__(self) -> None:
        self.host = "mikro"
        self.command = "bash run-debug.sh"
        self.repo = "https://github.com/arkitektio/mikro-server-next"
        self.scopes = {
            "mikro_read": "Read image from the database",
            "mikro_write": "Write image to the database",
        }
        self.image = "jhnnsrs/mikro:next"
        self.mount_repo = False
        self.build_repo = False
        self.buckets = ["media", "zarr", "parquet"]
        self.secret_key = secrets.token_hex(16)

    def get_dependencies(self):
        return create_default_service_dependencies()

    def get_builder(self):
        return "arkitekt.generic"

    def preflight(self, init: InitContext):
        for key, value in init.kwargs.items():
            setattr(self, key, value)

        self.service = create_default_service_yaml(init, self)

        self.initialized = True

    def build(self, context: ExecutionContext):
        context.docker_compose.set_nested("services", self.host, self.service)

    def get_options(self):
        with_repo = Option(
            subcommand="with_repo",
            help="The fakts url for connection",
            default=self.repo,
        )
        with_command = Option(
            subcommand="command",
            help="The fakts url for connection",
            default=self.command,
        )
        mount_repo = Option(
            subcommand="mount_repo",
            help="The fakts url for connection",
            type=bool,
            default=self.mount_repo,
        )
        build_repo = Option(
            subcommand="build_repo",
            help="The fakts url for connection",
            type=bool,
            default=self.build_repo,
        )
        with_host = Option(
            subcommand="host",
            help="The fakts url for connection",
            default=self.host,
        )
        with_secret_key = Option(
            subcommand="secret_key",
            help="The fakts url for connection",
            default=self.secret_key,
        )

        return [
            with_repo,
            mount_repo,
            build_repo,
            with_host,
            with_command,
            with_secret_key,
        ]





from typing import Dict, Any, Protocol
from blok import blok, InitContext, Option
from blok import service
from dataclasses import dataclass

@dataclass
class LivekitCredentials:
    api_key: str
    api_secret: str
    api_url: str

@service("io.livekit.livekit")
class LivekitService(Protocol):

    def retrieve_access(self) -> LivekitCredentials:
        ...




from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import YamlFile, Repo
from dataclasses import dataclass
from typing import Dict, Any, Protocol, Optional

from blok import blok, InitContext, service
from dataclasses import dataclass


@dataclass
class RedisConnection:
    host: str
    port: int
    dependency: Optional[str] = None


@blok("live.arkitekt.redis")
class RedisService(Protocol):
    def register(self) -> RedisConnection: ...



import rich_click as click
import subprocess
from .utils import search_username_in_docker_info
from arkitekt_next.cli.vars import get_console
from arkitekt_next.cli.types import Build
from rich.panel import Panel
from arkitekt_next.cli.io import get_builds, get_deployments, generate_deployment
from click import Context
import uuid


def check_if_build_already_deployed(build: Build) -> None:
    """Checks if a manifest has already been deployed. If it has, it raises a click.ClickException.

    Parameters
    ----------
    manifest : Manifest
        THe manifest to check

    Raises
    ------
    click.ClickException
        A click exception if the manifest has already been deployed
    """
    config = get_deployments()
    for deployment in config.deployments:
        if (
            deployment.manifest.identifier == build.manifest.identifier
            and deployment.manifest.version == build.manifest.version
            and deployment.flavour == build.flavour
        ):
            raise click.ClickException(
                f"Deployment of {build.manifest.identifier}/{build.manifest.version} in the {build.flavour} flavour already exists."
                + " You cannot deploy a build twice for the same version and flavour"
            )


@click.command()
@click.option("--build", help="The build run to use", type=str, default=None)
@click.option("--tag", help="The tag to use")
@click.pass_context
def publish(ctx: Context, build: str, tag: str) -> None:
    """Deploys aa previous build to dockerhub"""

    console = get_console(ctx)

    deployment_run = str(uuid.uuid4())

    builds = get_builds(selected_run=build)

    if len(builds) == 0:
        raise click.ClickException("Could not find any builds")

    docker_info = subprocess.check_output(["docker", "info"]).decode("utf-8")
    username = search_username_in_docker_info(docker_info)
    if not username:
        username = click.prompt(
            "Could not find username in docker info. Please provide your docker username"
        )

    for build_id, build_model in builds.items():
        if build_model.manifest.version == "dev":
            raise click.ClickException(
                "You cannot deploy a dev version. Please run `arkitekt_next version` first to set a version"
            )

        check_if_build_already_deployed(build_model)

        tag = tag or click.prompt(
            "The tag to use",
            default=f"{username}/{build_model.manifest.identifier}:{build_model.manifest.version}-{build_model.flavour}",
        )

        md = Panel("Building Docker Container")
        console.print(md)

        docker_run = subprocess.run(["docker", "tag", build_model.build_id, tag])
        if docker_run.returncode != 0:
            raise click.ClickException("Could not retag docker container")

        console.print(md)
        docker_run = subprocess.run(["docker", "push", tag])
        if docker_run.returncode != 0:
            raise click.ClickException("Could not push docker container")

        generate_deployment(
            deployment_run,
            build_model,
            tag,
            definitions=[],
        )

        md = Panel(
            f"[bold green] Sucessfully pushed {tag} your plugin to dockerhub"
            + "We have also generated a deployment file for you. "
            + "Make sure to commit and push your changes to github to make them available to others.",
            title="Yeah! Success",
            title_align="center",
            border_style="green",
            style="green",
        )
        console.print(md)



from typing import Optional

from arkitekt_next.bloks.services.db import DBService, DBCredentials
from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import YamlFile, Repo
from pydantic import BaseModel
import pydantic
import namegenerator
import secrets
from blok import blok, InitContext




@blok(DBService)
class PostgresBlok(BaseModel):
    host: str = "db"
    port: int = 5432
    skip: bool = False
    password: str = pydantic.Field(default_factory=lambda: secrets.token_hex(16))
    user: str = pydantic.Field(default_factory=lambda: namegenerator.gen(separator=""))
    image: str = "jhnnsrs/daten:next"

    registered_dbs: dict[str, DBCredentials] = {}

    def get_dependencies(self):
        return []

    def get_identifier(self):
        return "live.arkitekt.postgres"

    def register_db(self, db_name: str) -> DBCredentials:
        if db_name in self.registered_dbs:
            return self.registered_dbs[db_name]
        else:
            access_credentials = DBCredentials(
                password=self.password,
                username=self.user,
                host=self.host,
                port=self.port,
                db_name=db_name,
                dependency=self.host if not self.skip else None,
            )
            self.registered_dbs[db_name] = access_credentials
            return access_credentials

    def preflight(self, init: InitContext):
        for key, value in init.kwargs.items():
            setattr(self, key, value)

    def build(self, context: ExecutionContext):
        db_service = {
            "environment": {
                "POSTGRES_USER": self.user,
                "POSTGRES_PASSWORD": self.password,
                "POSTGRES_MULTIPLE_DATABASES": ",".join(self.registered_dbs.keys()),
            },
            "image": self.image,
            "labels": ["fakts.service=live.arkitekt.postgres"],
        }

        context.docker_compose.set_nested(f"services", self.host, db_service)

    def get_options(self):
        with_postgres_password = Option(
            subcommand="password",
            help="The postgres password for connection",
            default=self.password,
        )
        with_user_password = Option(
            subcommand="user",
            help="The postgress user_name",
            default=self.user,
        )
        skip_build = Option(
            subcommand="skip",
            help="Should the service not be created? E.g when pointing outwards?",
            default=self.skip,
        )
        with_image = Option(
            subcommand="image",
            help="The image to use for the service",
            default=self.image,
        )

        return [with_postgres_password, skip_build, with_user_password, with_image]



import click

from pydantic import BaseModel
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend as crypto_default_backend
from typing import Dict
from arkitekt_next.bloks.secret import SecretBlok
from arkitekt_next.bloks.services.admin import AdminService
from arkitekt_next.bloks.services.db import DBService
from arkitekt_next.bloks.services.gateway import GatewayService
from arkitekt_next.bloks.services.livekit import LivekitService
from arkitekt_next.bloks.services.lok import LokCredentials, LokService
import yaml
import secrets
from dataclasses import asdict

from arkitekt_next.bloks.services.redis import RedisService
from blok import blok, InitContext, ExecutionContext, Option
from blok.bloks.services.dns import DnsService
from blok.tree import YamlFile, Repo
from blok import blok, InitContext


DEFAULT_ARKITEKT_URL = "http://localhost:8000"


# Define a custom user type that will parse and validate the user input
class UserParamType(click.ParamType):
    name = "user"

    def convert(self, value, param, ctx):
        if isinstance(value, dict):
            return value
        try:
            name, password = value.split(":")
            return {"name": name, "password": password}
        except ValueError:
            self.fail(
                f"User '{value}' is not in the correct format. It should be 'name:password'.",
                param,
                ctx,
            )


USER = UserParamType()


# Define a custom user type that will parse and validate the user input
class GroupParamType(click.ParamType):
    name = "group"

    def convert(self, value, param, ctx):
        if isinstance(value, dict):
            return value
        try:
            name, description = value.split(":")
            return {"name": name, "description": description}
        except ValueError:
            self.fail(
                f"User '{value}' is not in the correct format. It should be 'name:password'.",
                param,
                ctx,
            )


GROUP = GroupParamType()


class RedeemTokenParamType(click.ParamType):
    name = "redeem_token"

    def convert(self, value, param, ctx):
        if isinstance(value, dict):
            assert "user" in value, f"scope is required {value}"
            assert "token" in value, f"description is required {value}"
            return value

        try:
            user, token = value.split(":")
            return {"user": user, "token": token}
        except ValueError:
            self.fail(
                f"RedeemToken '{value}' is not in the correct format. It should be 'username:token'.",
                param,
                ctx,
            )


TOKEN = RedeemTokenParamType()


class ScopeParamType(click.ParamType):
    name = "scope"

    def convert(self, value, param, ctx):
        if isinstance(value, dict):
            assert "scope" in value, f"scope is required {value}"
            assert "description" in value, f"description is required {value}"
            return value

        try:
            name, description = value.split(":")
            return {"scope": name, "description": description}
        except ValueError:
            self.fail(
                f"Scopes '{value}' is not in the correct format. It should be 'scope:description'.",
                param,
                ctx,
            )


SCOPE = ScopeParamType()


@blok(LokService)
class LokBlok:
    db_name: str

    def __init__(self) -> None:
        self.db_name = "lok_db"
        self.mount_repo = False
        self.build_repo = False
        self.private_key = None
        self.public_key = None
        self.host = "lok"
        self.with_repo = False
        self.command = "bash run-debug.sh"
        self.repo = "https://github.com/jhnnsrs/lok-server-next"
        self.image = "jhnnsrs/lok:next"
        self.users = []
        self.tokens = []
        self.groups = []
        self.secret_key = secrets.token_hex(16)
        self.scopes = {
            "openid": "The open id connect scope",
            "read": "A generic read access",
            "write": "A generic write access",
        }
        self.key = None
        self.deployment_name = "default"
        self.token_expiry_seconds = 700000
        self.preformed_redeem_tokens = [secrets.token_hex(16) for i in range(80)]
        self.registered_tokens = {}

    def retrieve_credentials(self) -> LokCredentials:
        return LokCredentials(
            public_key=self.public_key, key_type="RS256", issuer="lok"
        )

    def retrieve_labels(self, service_name: str, builder_name: str) -> list[str]:
        return [
            f"fakts.service={service_name}",
            f"fakts.builder={builder_name}",
        ]

    def retrieve_token(self, user: str = "admin") -> str:
        new_token = self.secret_blok.retrieve_secret()
        self.registered_tokens[user] = new_token

        return new_token

    def register_scopes(self, scopes_dict: Dict[str, str]) -> LokCredentials:
        self.scopes = self.scopes | scopes_dict

    def preflight(
        self,
        init: InitContext,
        gateway: GatewayService,
        db: DBService,
        redis: RedisService,
        admin: AdminService,
        secrets: SecretBlok,
        livekit: LivekitService,
        dns: DnsService,
    ):
        for key, value in init.kwargs.items():
            setattr(self, key, value)

        assert self.public_key, "Public key is required"
        assert self.private_key, "Private key is required"

        gateway.expose("lok", 80, self.host, strip_prefix=False)
        gateway.expose_mapped(".well-known", 80, self.host, "lok")

        self.secret_blok = secrets
        self.postgress_access = db.register_db(self.host)
        self.redis_access = redis.register()
        self.admin_access = admin.retrieve()
        self.local_access = livekit.retrieve_access()
        self.dns_result = dns.get_dns_result()
        self.initialized = True

    def build(self, context: ExecutionContext):
        depends_on = []

        if self.redis_access.dependency:
            depends_on.append(self.redis_access.dependency)

        if self.postgress_access.dependency:
            depends_on.append(self.postgress_access.dependency)

        db_service = {
            "labels": ["fakts.service=live.arkitekt.lok", "fakts.builder=arkitekt.lok"],
            "depends_on": depends_on,
            "volumes": [
                "/var/run/docker.sock:/var/run/docker.sock",
                "./configs/lok.yaml:/workspace/config.yaml",
                "./certs:/certs",
            ],
        }

        if self.mount_repo:
            context.file_tree.set_nested("mounts", "lok", Repo(self.repo))
            db_service["volumes"].append("./mounts/lok:/lok")

        if self.build_repo:
            context.file_tree.set_nested("mounts", "lok", Repo(self.repo))
            db_service["build"] = "./mounts/lok"
        else:
            db_service["image"] = self.image

        db_service["command"] = self.command

        trusted_origins = []

        for i in self.dns_result.hostnames:
            trusted_origins.append(f"http://{i}")
            trusted_origins.append(f"https://{i}")

        configuration = YamlFile(
            **{
                "db": asdict(self.postgress_access),
                "users": [user for user in self.users],
                "django": {
                    "admin": asdict(self.admin_access),
                    "debug": True,
                    "hosts": ["*"],
                    "secret_key": self.secret_key,
                },
                "redis": asdict(self.redis_access),
                "lok": asdict(self.retrieve_credentials()),
                "private_key": self.private_key,
                "public_key": self.public_key,
                "scopes": self.scopes,
                "redeem_tokens": [
                    {"user": name, "token": token}
                    for name, token in self.registered_tokens.items()
                ],
                "groups": [group for group in self.groups],
                "deployment": {"name": self.deployment_name},
                "livekit": asdict(self.local_access),
                "token_expire_seconds": self.token_expiry_seconds,
                "apps": [],
                "force_script_name": "lok",
                "csrf_trusted_origins": trusted_origins,
            }
        )

        context.file_tree.set_nested("configs", "lok.yaml", configuration)

        context.docker_compose.set_nested("services", self.host, db_service)

    def get_options(self):
        key = rsa.generate_private_key(
            public_exponent=65537, key_size=2048, backend=crypto_default_backend()
        )

        private_key = key.private_bytes(
            crypto_serialization.Encoding.PEM,
            crypto_serialization.PrivateFormat.PKCS8,
            crypto_serialization.NoEncryption(),
        ).decode()

        public_key = (
            key.public_key()
            .public_bytes(
                crypto_serialization.Encoding.OpenSSH,
                crypto_serialization.PublicFormat.OpenSSH,
            )
            .decode()
        )

        with_fakts_url = Option(
            subcommand="db_name",
            help="The name of the database",
            default="db_name",
            show_default=True,
        )
        with_users = Option(
            subcommand="users",
            help="Users that should be greated by default. Format is name:password",
            default=["admin:admin"],
            multiple=True,
            type=USER,
            show_default=True,
        )
        with_groups = Option(
            subcommand="groups",
            help="Groups that should be greated by default. Format is name:description",
            default=["admin:admin_group"],
            multiple=True,
            type=GROUP,
            show_default=True,
        )
        with_repo = Option(
            subcommand="with_repo",
            help="Which repo should we use when building the service? Only active if build_repo or mount_repo is active",
            default=self.repo,
            show_default=True,
        )
        with_repo = Option(
            subcommand="command",
            help="Which command should be run when starting the service?",
            default=self.command,
            show_default=True,
        )
        mount_repo = Option(
            subcommand="mount_repo",
            help="The fakts url for connection",
            type=bool,
            default=False,
        )
        build_repo = Option(
            subcommand="build_repo",
            help="Should we build the container from the repo?",
            type=bool,
            default=False,
            show_default=True,
        )
        with_host = Option(
            subcommand="host",
            help="Which internal hostname should be used",
            default=self.host,
            show_default=True,
        )
        #
        with_public_key = Option(
            subcommand="public_key",
            help="The public key for the JWT creation",
            default=public_key,
            required=True,
            callback=validate_public_key,
        )
        with_private_key = Option(
            subcommand="private_key",
            help="The corresponding private key for the JWT creation",
            default=private_key,
            callback=validate_private_key,
            required=True,
        )
        with_secret_key = Option(
            subcommand="secret_key",
            help="The secret key to use for the django service",
            default=self.secret_key,
        )

        return [
            with_fakts_url,
            with_users,
            with_repo,
            mount_repo,
            with_groups,
            build_repo,
            with_host,
            with_private_key,
            with_public_key,
            with_secret_key,
        ]


def validate_public_key(ctx, param, value):
    if not value.startswith("ssh-rsa"):
        raise click.BadParameter(
            f"Public key must be in ssh-rsa format. Started with {value}"
        )
    return value


def validate_private_key(ctx, param, value):
    if not value.startswith("-----BEGIN PRIVATE KEY-----"):
        raise click.BadParameter(
            f"Private key must be in PEM format. Started with {value}"
        )
    return value



from .types import EasyApp
from typing import Optional, List


def build_arkitekt_app(
    manifest: Manifest,
    url=None,
    no_cache=False,
    headless=False,
    instance_id=None,
    token=None,
    app_kind="development",
    redeem_token=None,
):
    if redeem_token:
        fakts = build_arkitekt_next_redeem_fakts_next(
            manifest=manifest,
            redeem_token=redeem_token,
            url=url,
            no_cache=no_cache,
            headless=headless,
        )
    else:
        fakts = build_arkitekt_next_fakts_next(
            manifest=manifest,
            url=url,
            no_cache=no_cache,
            headless=headless,
            client_kind=app_kind,
        )

    herre = build_arkitekt_next_herre(fakts=fakts)

    try:
        from arkitekt_next_next.apps.service.rekuest_next import build_arkitekt_next_rekuest_next

        rekuest = build_arkitekt_next_rekuest_next(
            fakts=fakts, herre=herre, instance_id=instance_id
        )
    except ImportError as e:
        rekuest = ImportException(import_exception=e, install_library="rekuest_next")

    try:
        from arkitekt_next_next.apps.service.mikro_next import build_arkitekt_next_mikro_next

        mikro = build_arkitekt_next_mikro_next(fakts=fakts, herre=herre)
    except ImportError as e:
        raise e
        mikro = ImportException(import_exception=e, install_library="mikro_next")

    try:
        from arkitekt_next_next.apps.service.fluss_next import build_arkitekt_next_fluss

        fluss = build_arkitekt_next_fluss(herre=herre, fakts=fakts)
    except ImportError as e:
        raise e
        fluss = ImportException(import_exception=e, install_library="fluss_next")

    try:
        from arkitekt_next_next.apps.service.unlok_next import build_arkitekt_next_unlok_next

        unlok = build_arkitekt_next_unlok_next(herre=herre, fakts=fakts)
    except ImportError as e:
        raise e
        fluss = ImportException(import_exception=e, install_library="fluss_next")

    try:
        from arkitekt_next_next.apps.service.omero_ark import build_arkitekt_next_omero_ark

        omero_ark = build_arkitekt_next_omero_ark(herre=herre, fakts=fakts)
    except ImportError as e:
        omero_ark = ImportException(import_exception=e, install_library="omero_ark")

    try:
        from arkitekt_next_next.apps.service.kluster import build_arkitekt_next_kluster

        kluster = build_arkitekt_next_kluster(herre=herre, fakts=fakts)
    except ImportError as e:
        kluster = ImportException(import_exception=e, install_library="kluster")

    try:
        from arkitekt_next_next.apps.service.kabinet import build_arkitekt_next_kabinet

        kabinet = build_arkitekt_next_kabinet(herre=herre, fakts=fakts)
    except ImportError as e:
        kabinet = ImportException(import_exception=e, install_library="kluster")

    return NextApp(
        manifest=manifest,
        fakts=fakts,
        herre=herre,
        rekuest=rekuest,
        mikro=mikro,
        kabinet=kabinet,
        unlok=unlok,
        fluss=fluss,
        kluster=kluster,
        omero_ark=omero_ark,
    )



from typing import Dict, Any, Protocol
from blok import blok, InitContext, Option, ExecutionContext, service
from blok.tree import YamlFile




@service("live.arkitekt.mount")
class MountService(Protocol):

    def register_mount(self, name: str, file: YamlFile) -> str:
        ...


   



from blok import blok, InitContext, Option
from blok.tree import YamlFile, Repo
from typing import Any, Protocol
from blok.utils import check_protocol_compliance
from dataclasses import asdict
from arkitekt_next.bloks.services import (
    GatewayService,
    DBService,
    RedisService,
    S3Service,
    ConfigService,
    MountService,
    AdminService,
    SecretService,
    LokService,
)

from blok.bloks.services.dns import DnsService


class DefaultService(Protocol):
    service_name: str
    host: str
    buckets: list[str]
    scopes: dict[str, str]
    secret_key: str
    mount_repo: bool
    build_repo: bool

    def get_identifier(self) -> str: ...

    def get_builder(self) -> str: ...


def create_default_service_dependencies():
    return [
        DnsService,
        GatewayService,
        DBService,
        RedisService,
        S3Service,
        ConfigService,
        MountService,
        AdminService,
        SecretService,
        LokService,
    ]


def create_default_service_yaml(
    init: InitContext,
    self: DefaultService,
    additional_keys: dict[str, Any] = None,
) -> YamlFile:
    check_protocol_compliance(self, DefaultService)
    deps = init.dependencies
    additional_keys = additional_keys or {}

    init.get_service(LokService).register_scopes(self.scopes)

    path_name = self.host

    gateway_access = init.get_service(GatewayService).expose(path_name, 80, self.host)

    postgress_access = init.get_service(DBService).register_db(self.host)
    redis_access = init.get_service(RedisService).register()
    lok_access = init.get_service(LokService).retrieve_credentials()
    admin_access = init.get_service(AdminService).retrieve()
    minio_access = init.get_service(S3Service).create_buckets(self.buckets)
    lok_labels = init.get_service(LokService).retrieve_labels(
        self.get_identifier(), self.get_builder()
    )

    dns_result = init.get_service(DnsService).get_dns_result()

    csrf_trusted_origins = []
    for hostname in dns_result.hostnames:
        csrf_trusted_origins.append(f"http://{hostname}")
        csrf_trusted_origins.append(f"https://{hostname}")

    configuration = YamlFile(
        **{
            "db": asdict(postgress_access),
            "django": {
                "admin": asdict(admin_access),
                "debug": True,
                "hosts": ["*"],
                "secret_key": self.secret_key,
            },
            "redis": asdict(redis_access),
            "lok": asdict(lok_access),
            "s3": asdict(minio_access),
            "scopes": self.scopes,
            "force_script_name": path_name,
            "csrf_trusted_origins": csrf_trusted_origins,
            **additional_keys,
        }
    )

    config_mount = init.get_service(ConfigService).register_config(
        f"{self.host}.yaml", configuration
    )

    depends_on = []

    if redis_access.dependency:
        depends_on.append(redis_access.dependency)

    if postgress_access.dependency:
        depends_on.append(postgress_access.dependency)

    if minio_access.dependency:
        depends_on.append(minio_access.dependency)

    service = {
        "labels": lok_labels,
        "volumes": [f"{config_mount}:/workspace/config.yaml"],
        "depends_on": depends_on,
    }

    if self.mount_repo:
        mount = init.get_service(MountService).register_mount(
            self.host, Repo(self.repo)
        )
        service["volumes"].extend([f"{mount}:/workspace"])

    if self.build_repo:
        mount = init.get_service(MountService).register_mount(
            self.host, Repo(self.repo)
        )
        service["build"] = mount
    else:
        service["image"] = self.image

    service["command"] = self.command

    return service



from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import YamlFile, Repo
from dataclasses import dataclass
from typing import Dict, Any, Protocol

from blok import blok, InitContext, service


@blok("live.arkitekt.gateway")
class GatewayService(Protocol):
    def expose(
        self, path_name: str, port: int, host: str, strip_prefix: bool = True
    ): ...

    def retrieve_gateway_network(self): ...

    def expose_mapped(self, path_name: str, port: int, host: str, to_name: str): ...

    def expose_default(self, port: int, host: str): ...

    def expose_port(self, port: int, host: str, tls: bool = False): ...

    def expose_port_to(self, port: int, host: str, to_port: str, tls: bool = False): ...

    def get_internal_host(self): ...

    def get_https_port(self): ...

    def get_http_port(self): ...



from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import YamlFile, Repo
from dataclasses import dataclass
from typing import Dict, Any, Protocol

from blok import blok, InitContext, service


@blok("live.arkitekt.docker_socket")
class DockerSocketService(Protocol):

    def register_socket(self, name: str) -> str:
        self.registered_configs[name] = name
        return self.docker_socket

    



""" Utility functions for the Qt UI. """

import os

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
ASSETS_DIR = os.path.join(THIS_DIR, "assets")


def get_image_path(
    image_name: str, dark_mode: bool = False, assets_dir: str = ASSETS_DIR
) -> str:
    """Get the path to an image.

    Parameters
    ----------
    image_name : str
        Name of the image.
    dark_mode : bool, optional
        Whether to use the dark mode image, by default False.
    assets_dir : str, optional
        Directory where the assets are stored, by default ASSETS_DIR.

    Returns
    -------
    str
        Path to the image.
    """
    if dark_mode:
        return os.path.join(assets_dir, f"dark/{image_name}")
    return os.path.join(assets_dir, f"light/{image_name}")



import click

from arkitekt_next.bloks.services.redis import RedisService, RedisConnection
from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import YamlFile, Repo
from pydantic import BaseModel
from typing import Dict, Any, Optional

from blok import blok, InitContext



@blok(RedisService)
class RedisBlok:
    def __init__(self) -> None:
        self.host = "redis"
        self.port = 6379
        self.skip = False
        self.image = "redis"

    def get_identifier(self):
        return "live.arkitekt.redis"

    def get_dependencies(self):
        return []

    def register(self) -> RedisConnection:
        return RedisConnection(
            host=self.host,
            port=self.port,
            dependency=self.host if not self.skip else None,
        )

    def preflight(self, init: InitContext):
        for key, value in init.kwargs.items():
            setattr(self, key, value)

    def build(self, context: ExecutionContext):
        redis_service = {
            "environment": {
                "REDIS_HOST": self.host,
                "REDIS_PORT": self.port,
            },
            "image": self.image,
            "ports": [f"{self.port}:{self.port}"],
        }

        context.docker_compose.set_nested(f"services", self.host, redis_service)

    def get_options(self):
        with_port = Option(
            subcommand="port",
            help="Which port to use",
            type=int,
            default=self.port,
            show_default=True,
        )
        with_host = Option(
            subcommand="host",
            help="Which public hosts to use",
            type=str,
            default=self.host,
            show_default=True,
        )
        with_skip = Option(
            subcommand="skip",
            help="Skip docker creation (if using external redis?)",
           type=bool,
            default=self.skip,
        )
        with_image = Option(
            subcommand="image",
            help="The image to use for the service",
            default=self.image,
        )

        return [with_port, with_host, with_skip, with_image]



from pydantic import BaseModel
from typing import Dict, Any
from arkitekt_next.bloks.tailscale import TailscaleBlok
from blok import blok, InitContext, Renderer, Panel
from .livekit import LocalLiveKitBlok
from .mikro import MikroBlok
from .kabinet import KabinetBlok
from .rekuest import RekuestBlok
from .fluss import FlussBlok
from .gateway import GatewayBlok
from .internal_docker import InternalDockerBlok


class AdminCredentials(BaseModel):
    password: str
    username: str
    email: str


@blok("live.arkitekt")
class ArkitektBlok:

    def entry(self, renderer: Renderer):
        renderer.render(
            Panel(
                f"This is the arkitekt build that allows you to setup a full stack arkitekt application. Make sure to understand how bloks work before proceeding.",
                expand=False,
                title="Welcome to Arkitekt!",
                style="bold magenta",
            )
        )

    def preflight(self, lok: LocalLiveKitBlok, mikro: MikroBlok, kabinet: KabinetBlok, rekuest: RekuestBlok, fluss: FlussBlok, gateway: GatewayBlok, internal_engine: InternalDockerBlok, scale: TailscaleBlok):
        print (lok, mikro, kabinet, rekuest, fluss, gateway, internal_engine)


    def build(self, cwd):
        pass




from arkitekt_next.bloks.admin import AdminBlok
from arkitekt_next.bloks.arkitekt import ArkitektBlok
from arkitekt_next.bloks.mikro import MikroBlok
from arkitekt_next.bloks.fluss import FlussBlok
from arkitekt_next.bloks.redis import RedisBlok
from arkitekt_next.bloks.gateway import GatewayBlok
from arkitekt_next.bloks.livekit import LocalLiveKitBlok
from arkitekt_next.bloks.postgres import PostgresBlok
from arkitekt_next.bloks.minio import MinioBlok
from arkitekt_next.bloks.kabinet import KabinetBlok
from arkitekt_next.bloks.lok import LokBlok
from arkitekt_next.bloks.config import ConfigBlok
from arkitekt_next.bloks.mount import MountBlok
from arkitekt_next.bloks.internal_docker import InternalDockerBlok
from arkitekt_next.bloks.socket import DockerSocketBlok
from arkitekt_next.bloks.rekuest import RekuestBlok
from arkitekt_next.bloks.tailscale import TailscaleBlok
from arkitekt_next.bloks.secret import SecretBlok
from arkitekt_next.bloks.namegen import PreformedNamesBlok


def get_bloks():
    return [
        AdminBlok(),
        ArkitektBlok(),
        MikroBlok(),
        FlussBlok(),
        RedisBlok(),
        GatewayBlok(),
        LocalLiveKitBlok(),
        PostgresBlok(),
        MinioBlok(),
        LokBlok(),
        KabinetBlok(),
        MountBlok(),
        ConfigBlok(),
        InternalDockerBlok(),
        DockerSocketBlok(),
        RekuestBlok(),
        TailscaleBlok(),
        SecretBlok(),
        PreformedNamesBlok(),
    ]



from typing import Dict, Any
import secrets

from arkitekt_next.bloks.funcs import (
    create_default_service_yaml,
    create_default_service_dependencies,
)
from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import Repo, YamlFile


@blok("live.arkitekt.kabinet")
class KabinetBlok:
    def __init__(self) -> None:
        self.host = "kabinet"
        self.command = "bash run-debug.sh"
        self.repo = "https://github.com/jhnnsrs/kabinet-server"
        self.scopes = {
            "kabinet_deploy": "Deploy containers",
            "kabinet_add_repo": "Add repositories to the database",
        }
        self.mount_repo = False
        self.build_repo = False
        self.buckets = ["media"]
        self.secret_key = secrets.token_hex(16)
        self.ensured_repos = ["jhnnsrs/ome:main", "jhnnsrs/renderer:main"]
        self.image = "jhnnsrs/kabinet:next"

    def get_builder(self):
        return "arkitekt.generic"

    def get_dependencies(self):
        return create_default_service_dependencies()

    def preflight(self, init: InitContext):
        for key, value in init.kwargs.items():
            setattr(self, key, value)

        self.service = create_default_service_yaml(
            init, self, {"ensured_repos": self.ensured_repos}
        )

        self.initialized = True

    def build(self, context: ExecutionContext):
        context.docker_compose.set_nested("services", self.host, self.service)

    def get_options(self):
        with_repo = Option(
            subcommand="with_repo",
            help="Which repo should we use when building the service? Only active if build_repo or mount_repo is active",
            default=self.repo,
        )
        with_command = Option(
            subcommand="command",
            help="Which command should be run when starting the service",
            default=self.command,
        )
        mount_repo = Option(
            subcommand="mount_repo",
            help="Should we mount the repo into the container?",
            type=bool,
            default=self.mount_repo,
        )
        build_repo = Option(
            subcommand="build_repo",
            help="Should we build the container from the repo?",
            type=bool,
            default=self.build_repo,
        )
        with_host = Option(
            subcommand="host",
            help="How should the service be named inside the docker-compose file?",
            default=self.host,
        )
        with_secret_key = Option(
            subcommand="secret_key",
            help="The secret key to use for the django service",
            default=self.secret_key,
        )
        with_repos = Option(
            subcommand="repos",
            help="The default repos to enable for the service",
            default=self.secret_key,
        )

        return [
            with_repo,
            mount_repo,
            build_repo,
            with_host,
            with_command,
            with_secret_key,
            with_repos,
        ]



from arkitekt_next_next.apps.fallbacks import ImportException
from arkitekt_next_next.apps.service.fakts_next import (
    build_arkitekt_next_fakts_next,
    build_arkitekt_next_redeem_fakts_next,
)
from arkitekt_next_next.apps.service.herre import build_arkitekt_next_herre
from arkitekt_next_next.model import Manifest

from .types import NextApp


def build_next_app(
    manifest: Manifest,
    url=None,
    no_cache=False,
    headless=False,
    instance_id=None,
    token=None,
    app_kind="development",
    redeem_token=None,
):
    if redeem_token:
        fakts = build_arkitekt_next_redeem_fakts_next(
            manifest=manifest,
            redeem_token=redeem_token,
            url=url,
            no_cache=no_cache,
            headless=headless,
        )
    else:
        fakts = build_arkitekt_next_fakts_next(
            manifest=manifest,
            url=url,
            no_cache=no_cache,
            headless=headless,
            client_kind=app_kind,
        )

    herre = build_arkitekt_next_herre(fakts=fakts)




import semver
from .errors import ValidationError
import rich_click as click


def is_valid_semver(param: str, loaded=False) -> semver.VersionInfo:
    """Checks if the param is a valid semver version and returns it as a semver object"""
    if not semver.Version.is_valid(param):
        if loaded:
            raise click.ClickException(
                "Manifest version incorrect, please update your manifest to a valid semver version  [link=https://semver.org]semver[/link]."
            )

        raise ValidationError(
            "ArkitektNext versions need to follow semantic versioning. Please choose a correct format (examples: 0.0.0, 0.1.0, 0.0.0-alpha.1)"
        )
    return semver.Version.parse(param)



"""Qt related modules.

This module contains Modules that are Qt related, and help to integrate
ArkitektNext with Qt applications.

The main component is the MagicBar, which is a widget that can be added
to any Qt application. It will then allow the user to configure and connect
to ArkitektNext, and configure settings.
"""
from .magic_bar import MagicBar
from .builders import publicqt
from .types import *

__all__ = ["MagicBar", "publicqt", "QtApp"]



from arkitekt_next.apps.service.fakts_next import (
    build_arkitekt_next_redeem_fakts_next,
)
from arkitekt_next.apps.service.fakts_qt import build_arkitekt_next_qt_fakts
from arkitekt_next.apps.service.herre_qt import build_arkitekt_next_qt_herre
from arkitekt_next.utils import create_arkitekt_next_folder
from arkitekt_next.model import Manifest
from arkitekt_next.apps.types import App
from arkitekt_next.service_registry import ServiceBuilderRegistry, check_and_import_services
from arkitekt_next.constants import DEFAULT_ARKITEKT_URL
from qtpy import QtWidgets, QtCore
from typing import List, Optional
import os
import logging


def publicqt(
    identifier: str,
    version: str = "latest",
    logo: Optional[str] = None,
    scopes: Optional[List[str]] = None,
    log_level: str = "ERROR",
    registry: Optional[ServiceBuilderRegistry] = None,
    parent: Optional[QtWidgets.QWidget] = None,
    beacon_widget: Optional[QtWidgets.QWidget] = None,
    login_widget: Optional[QtWidgets.QWidget] = None,
    settings: Optional[QtCore.QSettings] = None,
    **kwargs,
) -> App:
    """Public QtApp creation

    A simple way to create an Arkitekt app with a public grant (allowing users to sign
    in with the application ) utlizing a retrieve grant (necessating a previous configuration
    of the application on the server side)

    Args:
        identifier (str): The apps identifier
        version (str, optional): The apps verion. Defaults to "latest".
        parent (QtWidget, optional): The QtParent (for the login and server select widget). Defaults to None.

    Returns:
        Arkitekt: The Arkitekt app
    """

    registry = registry or check_and_import_services()

    settings = settings or QtCore.QSettings("arkitekt_next", f"{identifier}:{version}")


    manifest = Manifest(
        version=version,
        identifier=identifier,
        scopes=scopes if scopes else ["openid"],
        logo=logo,
        requirements=registry.get_requirements(),
    )

    fakts = build_arkitekt_next_qt_fakts(
        manifest=manifest,
        beacon_widget=beacon_widget,
        parent=parent,
        settings=settings,
    )

    herre = build_arkitekt_next_qt_herre(manifest, fakts=fakts, login_widget=login_widget, parent=parent, settings=settings)

    params = kwargs


    try:
        from rich.logging import RichHandler

        logging.basicConfig(level=log_level, handlers=[RichHandler()])
    except ImportError:
        logging.basicConfig(level=log_level)

    app = App(
        fakts=fakts,
        herre=herre,
        manifest=manifest,
        services=registry.build_service_map(fakts=fakts, herre=herre, params=params, manifest=manifest),
    )

    app.enter()

    return app
